/* Copyright (c) 2009-2010 Ondrej Donek.
   See the file LICENSE.txt for licensing information. */

/**
 * This serves all about storage. Should not be used by end developer. They
 * should rather use data objects prototypes instead (for example
 * tbNotesNotePrototype).
 * Each of these data objects provide same simple interface for storing data.
 * 
 * @link http://developer.mozilla.org/en/Extensions/Thunderbird/HowTos/Common_Thunderbird_Extension_Techniques/Use_SQLite
 *
 * @type Object
 */
var tbNotesDb =
{
	
	// Some 'global' variables
	__dbInitialized : false,
	__dirSrv : null,
	__dbFile : null,
	__dbSrv : null,
	__dbCon : null,
	
	// Object describing our database schema
	__dbSchema: {
		tables: {
			tbn_notes:
				"id INTEGER PRIMARY KEY, \
				 target_type TEXT NOT NULL DEFAULT '', \
				 target_key TEXT NOT NULL DEFAULT '', \
				 category_id INTEGER NOT NULL, \
				 created TEXT NOT NULL, \
				 updated TEXT NOT NULL, \
				 title TEXT NOT NULL, \
				 body TEXT, \
				 deleted INTEGER DEFAULT 0 ",
			tbn_categories:
				"id INTEGER PRIMARY KEY, \
				 parent_id INTEGER, \
				 name TEXT NOT NULL, \
				 description TEXT ",
			tbn_links:
				"id INTEGER PRIMARY KEY, \
				 note_id INTEGER, \
				 url TEXT "
		}
	},
	
	// Holds array of queries for updating db (used in development versions)
	__dbSchemaUpdates : {
		// from 0.1.1 to 0.1.5
		updateNotes: "ALTER TABLE tbn_notes ADD COLUMN target_name TEXT NOT NULL DEFAULT '' " ,
		updateCategories: "ALTER TABLE tbn_categories ADD COLUMN color TEXT NOT NULL DEFAULT '' ",
		// from 0.1.5 to 0.2
		updateLinks: "ALTER TABLE tbn_links ADD COLUMN title TEXT NOT NULL DEFAULT '' ",
		updateTargetTypes1: "UPDATE `tbn_notes` SET `target_type` = '' WHERE  `target_type` = 'none' ",
		updateTargetTypes2: "UPDATE `tbn_notes` SET `target_type` = 'mailmessage' WHERE  `target_type` = 'message' ", 
		// from 0.2.5 - this is because the bug that notes are missing when you move the message
		// to the another folder.
		updateNotesTarget: "ALTER TABLE tbn_notes ADD COLUMN target_id TEXT NOT NULL DEFAULT '' "
	},

	/**
	 * Returns TRUE if tbNotes'db part was successfully initialized.
	 * 
	 * @return Boolean
	 */
	isInitialized : function() {
		// And what about `__dbCon.connectionReady`
		return this.__dbInitialized;
	},
  
	/**
	 * Initialize database and establish the connection
	 * 
	 * @returns boolean
	 */
	init : function() 
	{
		try {
			this.__dirSrv = Components.classes["@mozilla.org/file/directory_service;1"].
					getService(Components.interfaces.nsIProperties);
			this.__dbFile = this.__dirSrv.get("ProfD", Components.interfaces.nsIFile);  
			this.__dbFile.append("tbnotes.sqlite");

			// Initialize db service and open connection
			this.__dbSrv = Components.classes["@mozilla.org/storage/service;1"].
					getService(Components.interfaces.mozIStorageService);
			this.__dbCon = this.__dbSrv.openDatabase(this.__dbFile);

			// Check if all tables exist
			for(var name in this.__dbSchema.tables) {
				if(!this.__dbCon.tableExists(name)) {
					this.__dbCon.createTable(name, this.__dbSchema.tables[name]);
				}
			}

			// Runs update commands
			for(var name in this.__dbSchemaUpdates) {
				try { 
					this.__dbCon.executeSimpleSQL(this.__dbSchemaUpdates[name]);
				} catch(e) {
					//Components.utils.reportError("Error occured when running update query '" + name + "'. " +
					//							 "Probably database is already updated.");
				}
			}
		} catch(e) {
			Components.utils.reportError(e);

			var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
					getService(Components.interfaces.nsIPromptService);
			var title = tbLinkDialog.mStrbundle.getString("errorDlgTitle");
			var msg = tbLinkDialog.mStrbundle.getString("errorDlgMsg");

			prompts.alert(window, title, msg);

			return false;
		}

		return false;
	},
	
	// ========================================================================
	// Notes
	
	/**
	 * Returns note by its id. Otherwise returns NULL.
	 * 
	 * @param aId {integer}
	 * @return {tbNotesNotePrototype|NULL}
	 */
	getNote : function(aId) {
		return this.getNoteById(aId);
	},
	
	// Alias for getNote
	getNoteById : function(aId) 
	{
		var query = "SELECT * " + 
		            "FROM tbn_notes " + 
					"WHERE id = :note_id " + 
					"LIMIT 1 ";
		var stmt = this.__dbCon.createStatement(query);
		stmt.params.note_id = aId;
		
		while(stmt.executeStep()) {
			if(stmt.row.id > 0) {
				return new tbNotesNotePrototype(stmt.row);
			}
		}
		
		return null;
	},
	
	/**
	 * Returns note which is attached to the given message. If there is 
	 * no such note returns NULL.
	 * 
	 * @param aMsg {object}
	 * @return {tbNotesNotePrototype|NULL}
	 */
	getNoteBySelectedMsg : function(aMsg) 
	{
		var query = "";
		var stmt  = null;
		
		if(typeof(aMsg.id) == "undefined" || aMsg.id == null) {
			var query = "SELECT * " + 
						"FROM tbn_notes " + 
						"WHERE " + 
						"  target_key = :target_key AND " + 
						"  deleted = 0 " + 
						"LIMIT 1 ";
			var stmt = this.__dbCon.createStatement(query);
			stmt.params.target_key = aMsg.key;
		} else {
			var query = "SELECT * " + 
						"FROM tbn_notes " + 
						"WHERE " + 
						"  ( target_key = :target_key OR target_id = :target_id ) AND " + 
						"  deleted = 0 " + 
						"LIMIT 1 ";
			var stmt = this.__dbCon.createStatement(query);
			stmt.params.target_key = aMsg.key;
			stmt.params.target_id = aMsg.id;
		}
		
		while(stmt.executeStep()) {
			if(stmt.row.id > 0) {
				return new tbNotesNotePrototype(stmt.row);
			}
		}
		
		return null;
	},
	
	/**
	 * Returns all notes attached to the given email message.
	 * 
	 * @param aMsg {object}
	 * @return {array} Array of instances of tbNotesNotePrototype.
	 */
	getNotesBySelectedMsg : function(aMsg)
	{
		var notes = new Array();
		var query = "";
		var stmt  = null;
		
		// [ondrejd] - fix from tbNotes 0.3.1
		if(aMsg == null) return notes;
		
		if(typeof(aMsg.id) == "undefined" || aMsg.id == null) {
			var query = "SELECT * " + 
						"FROM tbn_notes " + 
						"WHERE " + 
						"  target_key = :target_key AND " + 
						"  deleted = 0 ";
			var stmt = this.__dbCon.createStatement(query);
			stmt.params.target_key = aMsg.key;
		} else {
			var query = "SELECT * " + 
						"FROM tbn_notes " + 
						"WHERE " + 
						"  ( target_key = :target_key OR target_id = :target_id ) AND " + 
						"  deleted = 0 ";
			var stmt = this.__dbCon.createStatement(query);
			stmt.params.target_key = aMsg.key;
			stmt.params.target_id = aMsg.id;
		}
		
		while(stmt.executeStep()) {
			if(stmt.row.id > 0) {
				notes.push((new tbNotesNotePrototype(stmt.row)));
			}
		}
		
		return notes;
	},

	/** 
	 * Returns all notes as objects.
	 * 
	 * @param aAlsoDeleted {boolean}
	 * @return {array} Array of instances of tbNotesNotePrototype
	 */
	getNotes : function(aAlsoDeleted)
	{
		var notes = new Array();
		var query = "SELECT * FROM tbn_notes " + 
				((aAlsoDeleted !== true) ? "WHERE deleted = 0 " : "");


		// TODO Here is throwen exception that __dbCon is not defined!!! The strange is 
		//      that this error occures before the _init is loaded... So try the run 
		//      _init here if __dbCon is not defined.
		if(typeof(this.__dbCon) == "undefined" || this.__dbCon == null) {
			this.init();
		}

		var stmt = this.__dbCon.createStatement(query);

		while(stmt.executeStep()) {
			if(stmt.row.id > 0) {
				notes.push((new tbNotesNotePrototype(stmt.row)));
			}
		}

		return notes;
	},
	
	/**
	 * Adds new note into the database
	 * 
	 * @param aNote {tbNotesNotePrototype}
	 * @return {Integer|NULL}
	 */
	addNote : function(aNote) 
	{
		try {
			var query = 
					"INSERT INTO tbn_notes " + 
					"  (ID, target_type, target_key, target_name, target_id, category_id, " + 
					"   created, updated, title, body, deleted) " + 
					"VALUES " + 
					"  ( NULL , :target_type , :target_key , :target_name , :target_id , :category_id , " + 
					"    :created , :updated , :title , :body , :deleted ) ";
			
			var stmt = tbNotesDb.__dbCon.createStatement(query);
			// [ondrejd] - fix from tbNotes 0.3.1
			stmt.params.target_type = (aNote.mTargetType == null) ? '' : aNote.mTargetType;
			stmt.params.target_key  = aNote.mTargetKey;
			stmt.params.target_name = aNote.mTargetName;
			stmt.params.target_id = aNote.mTargetId;
			stmt.params.category_id = (aNote.mCategoryId == null) ? 0 : aNote.mCategoryId;
			stmt.params.created = tbNotesUtils.getCurrentDatetime();
			stmt.params.updated = tbNotesUtils.getCurrentDatetime();
			stmt.params.title = aNote.mTitle;
			stmt.params.body = aNote.mBody;
			stmt.params.deleted = (aNote.isMarkedAsDeleted() === true) ? 1 : 0;
			
			try {
				stmt.execute();
			} catch(e) {
				Components.utils.reportError("Database error: [" + tbNotesDb.__dbCon.lastError + "]: " + 
						tbNotesDb.__dbCon.lastErrorString);
				Components.utils.reportError(aNote.toString());
			}
		} catch(e) {
			Components.utils.reportError(e);
			return null;
		}
    
		return tbNotesDb.__dbCon.lastInsertRowID;
	},
  
	/**
	 * Removes note with given ID permanently from the database.
	 * 
	 * @param aId {Integer}
	 * @return {Boolean}
	 */
	removeNote : function(aId) 
	{
		Components.utils.reportError("Removing note with id: " + aId);
		try {
			var query = "DELETE FROM tbn_notes WHERE id = :note_id LIMIT 1 ";
			// TODO This throwing exception 'SQL error around LIMIT'.
			var stmt = this.__dbCon.createStatement(query);
			stmt.params.note_id = aId;
			stmt.execute();
		} catch(e) {
			Components.utils.reportError("Database error: [" + this.__dbCon.lastError + "]: " + this.__dbCon.lastErrorString);
			return false;
		}

		return true;
	},
	
	/**
	 * Updates note's record in the database
	 * 
	 * @param aNote {tbNotesNotePrototype}
	 * @return {Boolean}
	 */
	updateNote : function(aNote) 
	{
		try {
			// FIXME This query doesn't working!
			var query = 
					"UPDATE tbn_notes " + 
					"SET " + 
					"  target_type = :target_type , " + 
					"  target_key = :target_key , " + 
					"  target_name = :target_name , " + 
					"  updated = :updated , " + 
					"  title = :title ,  " + 
					"  body = :body , " +
					"  deleted = :deleted , " + 
					"  target_id = :target_id " + 
					"WHERE " + 
					"  id = :note_id ";
			
			var stmt = this.__dbCon.createStatement(query);
			stmt.params.target_type = aNote.mTargetType;
			stmt.params.target_key = aNote.mTargetKey;
			stmt.params.target_name = aNote.mTargetName;
			stmt.params.updated = tbNotesUtils.getCurrentDatetime();
			stmt.params.title = aNote.mTitle;
			stmt.params.body = aNote.mBody;
			stmt.params.note_id = aNote.mId;
			stmt.params.deleted = (aNote.isMarkedAsDeleted()) ? 1 : 0;
			stmt.params.target_id = aNote.mTargetId;
			stmt.execute();
		} catch(e) {
			Components.utils.reportError("Database error: [" + this.__dbCon.lastError + "]: " + 
					this.__dbCon.lastErrorString);
			return false;
		}
    
		return true;
	},
  
	/**
	 * Mark note with given ID as deleted.
	 * 
	 * @param aId {Integer}
	 * @return {Boolean}
	 */
	deleteNote : function(aId) 
	{
		try {
			var query = "UPDATE tbn_notes SET deleted = 1 WHERE id = :note_id ";
			var stmt = this.__dbCon.createStatement(query);
			stmt.params.note_id = aId;
			stmt.execute();
		} catch(e) {
			Components.utils.reportError("Database error: [" + this.__dbCon.lastError + "]: " + 
					this.__dbCon.lastErrorString);
			return false;
		}

		return true;
	},

	// ========================================================================
	// Categories
	
	// TODO Implement `getCategory` method!
	// TODO Implement `getCategories` method!
	// TODO Implement `addCategory` method!
	// TODO Implement `removeCategory` method!
	// TODO Implement `updateCategory` method!

	// ========================================================================
	// Links
	
	/**
	 * Returns link by its id. Otherwise returns NULL.
	 * 
	 * @param aId {integer}
	 * @return {tbNotesLinkPrototype|NULL}
	 */
	getLink : function(aId) {
		return this.getLinkById(aId);
	},
	
	// Alias for getLink
	getLinkById : function(aId) 
	{
		var query = "SELECT * FROM tbn_links WHERE id = :link_id ";
		var stmt = this.__dbCon.createStatement(query);
		stmt.params.link_id = aId;
		
		while(stmt.executeStep()) {
			if(stmt.row.id > 0) {
        return new tbNotesLinkPrototype(stmt.row);
			}
		}
		
		return null;
	},
  
  /** 
   * Returns all links as objects.
   * 
   * @return {array} Array of instances of tbNotesLinkPrototype
   */
  getLinks : function()
  {
    var links = new Array();
    var stmt = this.__dbCon.createStatement("SELECT * FROM tbn_links ");
    
    while(stmt.executeStep()) {
      if(stmt.row.id > 0) {
        links.push((new tbNotesLinkPrototype(stmt.row)));
      }
    }
    
    return links;
  },
  
  /** 
   * Returns all links which are attached to the note with given ID.
   *
   * @param aNoteId {integer}
   * @return {array} Array of instances of tbNotesLinkPrototype
   */
  getLinksByNoteId : function(aNoteId)
  {
    var links = new Array();
    var stmt = this.__dbCon.createStatement("SELECT * FROM tbn_links WHERE note_id = '" + aNoteId + "' ");
    
    while(stmt.executeStep()) {
      if(stmt.row.id > 0) {
        links.push((new tbNotesLinkPrototype(stmt.row)));
      }
    }
    
    return links;
  },
	
	/**
	 * Adds new link into the database
	 * 
	 * @param aLink {tbNotesLinkPrototype}
	 * @return {Integer|NULL}
	 */
	addLink : function(aLink) 
  {
		try {
			var query = 
					"INSERT INTO tbn_links (note_id, url, title) " + 
					"VALUES ( :note_id , :url , :title ); ";
			
			var stmt = this.__dbCon.createStatement(query);
			stmt.params.note_id = aLink.mNoteId;
			stmt.params.url = aLink.mUrl;
			stmt.params.title = aLink.mTitle;
			stmt.execute();
		} catch(e) {
			Components.utils.reportError("Database error: [" + this.__dbCon.lastError + "]: " + this.__dbCon.lastErrorString);
      return null;
		}
    
    return this.__dbCon.lastInsertRowID;
	},
  
  /**
   * Removes link with given ID permanently from the database.
   * 
   * @param aId {Integer}
   * @return {Boolean}
   */
  removeLink : function(aId) 
  {
    try {
			var query = "DELETE FROM tbn_links WHERE id = :link_id ";
			var stmt = this.__dbCon.createStatement(query);
			stmt.params.link_id = aId;
			stmt.execute();
    } catch(e) {
			Components.utils.reportError("Database error: [" + this.__dbCon.lastError + "]: " + this.__dbCon.lastErrorString);
			return false;
    }
    
    return true;
  },
	
	/**
	 * Updates link
	 * 
	 * @param aLink {tbNotesLinkPrototype}
	 * @return {Boolean}
	 */
	updateLink : function(aLink) 
  {
		try {
			var query = 
        "UPDATE tbn_links " + 
        "SET note_id = :note_id , url = :url , title = :title " + 
        "WHERE id = :link_id ";
			
			var stmt = this.__dbCon.createStatement(query);
			stmt.params.note_id = aLink.mNoteId;
			stmt.params.url = aLink.mUrl;
			stmt.params.title = aLink.mTitle;
			stmt.params.link_id = aLink.mId;
			stmt.execute();
		} catch(e) {
			Components.utils.reportError("Database error: [" + this.__dbCon.lastError + "]: " + this.__dbCon.lastErrorString);
      return false;
		}
    
    return true;
	}

}; // End of tbNotesDb
